/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*  This is a plugin intended to help with making panoramas.
 *        Copyright (C) 2003  Akkana Peck.
 *
 *  To build: gimptool --build pandora.c
 *  To install: gimptool --install-strip pandora.c
 */

/* Status: this runs the first time it's rebuilt;
 * on subsequent runs of gimp-1.2.4-pre2, gimp crashes doing:
 *  plug_in = g_new (PlugIn, 1);
 * on line 844 of plug_in.c.
 * I tried initializing PlugIn = 0 but it didn't help.
 */

#ifdef IN_GIMP_TREE
#include "config.h"
#include "libgimp/stdplugins-intl.h"
#else
/* gimptool doesn't understand these: */
#define _(s)   s
#define N_(s)  s
#define INIT_I18N() 
#define INIT_I18N_UI() 
#endif

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <libgen.h>

#ifdef __GNUC__
#warning GTK_DISABLE_DEPRECATED
#endif
#undef GTK_DISABLE_DEPRECATED

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <gtk/gtkfilesel.h>
#include <gtk/gtkclist.h>

/* Backwards compatibility defines for gimp 1.2: */
#ifndef G_N_ELEMENTS
#define GIMP_1_2 1
#define GimpRunMode gint32
#define drawable_id id
#define G_CALLBACK GTK_SIGNAL_FUNC
#define G_OBJECT GTK_OBJECT
#define g_signal_connect gtk_signal_connect
#define gtk_check_button_new_with_mnemonic gtk_check_button_new_with_label

#define GTK_STOCK_CANCEL _("Cancel")
#define GTK_STOCK_OK _("Ok")

#define G_N_ELEMENTS(a) ((sizeof a) / (sizeof (*(a))))
#endif /* G_N_ELEMENTS, gimp 1.2 backward compatibility */

#define MAXIMAGES 50

#define FILELIST_WIDTH   400
#define FILELIST_HEIGHT  250
#define SCALE_WIDTH       60
#define SPIN_WIDTH        60

typedef struct
{
  gint       run;
} PandoraInterface;

typedef struct
{
  gchar** files;
  gdouble overlap;
} PandoraVals;

static PandoraVals avals =
{
    0,   /* file list */
  .31    /* initial overlap fraction */
};

/* Declare local functions.
 */
static void      query (void);
static void      run   (gchar        *name,
                        gint          nparams,
                        GimpParam    *param,
                        gint         *nreturn_vals,
                        GimpParam   **return_vals);
static gint32    make_pandora ();

/*  user interface functions  */
static gint      pandora_dialog (void);

static PandoraInterface aint =
{
  FALSE,        /* run */
};

GimpPlugInInfo PLUG_IN_INFO =
{
  NULL,  /* init_proc  */
  NULL,  /* quit_proc  */
  query, /* query_proc */
  run,   /* run_proc   */
};

MAIN ()

static void
query (void)
{
  static GimpParamDef args[] =
  {
    { GIMP_PDB_INT32,       "run_mode", "Interactive, non-interactive" },
    { GIMP_PDB_STRINGARRAY, "files", "Input files" },
  };
  static GimpParamDef return_vals[] =
  {
    { GIMP_PDB_INT32, "Status",   "Success = 0" },
    { GIMP_PDB_INT32, "Pano Img id", "The new image." }
  }; 

#ifdef IN_GIMP_TREE
  INIT_I18N();
#endif

  gimp_install_procedure ("extension_pandora",
                          "Collect images together to make a panorama",
                          "Help not yet written for this plug-in",
                          "Akkana Peck",
                          "Akkana Peck",
                          "2003",
                          N_("<Toolbox>/Xtns/Pandora..."),
                          "",
                          GIMP_EXTENSION,
                          G_N_ELEMENTS (args), G_N_ELEMENTS (return_vals),
                          args, return_vals);
}

static void
run (gchar      *name,
     gint        nparams,
     GimpParam  *param,
     gint       *nreturn_vals,
     GimpParam **return_vals)
{
  static GimpParam values[2];
  GimpRunMode  run_mode;
  GimpPDBStatusType status = GIMP_PDB_SUCCESS;

  run_mode = param[0].data.d_int32;

  *nreturn_vals = 2;
  *return_vals  = values;
  values[0].type          = GIMP_PDB_STATUS;
  values[0].data.d_status = status;
  values[1].type = GIMP_PDB_INT32;
  values[1].data.d_image = -1;

  switch (run_mode)
  {
    case GIMP_RUN_INTERACTIVE:
      INIT_I18N_UI();
      /*  Possibly retrieve data  */
      gimp_get_data ("plug_in_pandora", &avals);

      /*  First acquire information with a dialog  */
      if (! pandora_dialog ())
	return;
      break;

    case GIMP_RUN_NONINTERACTIVE:
      INIT_I18N();
      /*  Make sure all the arguments are there!  */
      if (nparams != 15)
        status = GIMP_PDB_CALLING_ERROR;
      if (status == GIMP_PDB_SUCCESS)
      {
        avals.files = param[1].data.d_stringarray;
      }
      break;

    case GIMP_RUN_WITH_LAST_VALS:
      /*  Possibly retrieve data  */
      gimp_get_data ("plug_in_pandora", &avals);
      break;

    default:
      break;
    }

  /*  Do it!  */
  if (status == GIMP_PDB_SUCCESS)
  {
    /*  run the effect  */
    values[1].data.d_image = make_pandora ();

    /*  If the run mode is interactive, flush the displays  */
    if (run_mode != GIMP_RUN_NONINTERACTIVE)
      gimp_displays_flush ();

    /*  Store avals data  */
    if (run_mode == GIMP_RUN_INTERACTIVE)
      gimp_set_data ("plug_in_pandora", &avals, sizeof (PandoraVals));
  }
  else if (status == GIMP_PDB_SUCCESS)
  {
    /* gimp_message ("pandora: cannot operate on indexed color images"); */
    status = GIMP_PDB_EXECUTION_ERROR;
  }

  values[0].data.d_status = status;

  /*gimp_drawable_detach (drawable);*/
}

static GtkWidget *pandora_fsb = NULL;
static GtkWidget *pandora_filelist = NULL;

/* Add a file: this is called from the fsb's ok button, or from doubleclick */
static void
pandora_file_selection_ok (GtkWidget *widget,
                           gpointer   data)
{
  char* sarr[2];
  sarr[0] = gtk_file_selection_get_filename (GTK_FILE_SELECTION (data));
  sarr[1] = 0;
  gtk_clist_append(GTK_CLIST (pandora_filelist), sarr);
}

static void
pandora_file_selection_cancel (GtkWidget *widget,
                               gpointer   data)
{
  gtk_widget_hide (GTK_WIDGET (data));
}

static void
pandora_show_file_selection (GtkWidget *widget)
{
  if (!pandora_fsb)
    {
      pandora_fsb = gtk_file_selection_new (_("Panorama images"));

      /* Rename the OK and Cancel buttons */
#ifdef GIMP_1_2
      gtk_label_set_text( GTK_LABEL (GTK_BIN(GTK_FILE_SELECTION (pandora_fsb)->ok_button)->child), _("Add") );
      gtk_label_set_text( GTK_LABEL (GTK_BIN(GTK_FILE_SELECTION (pandora_fsb)->cancel_button)->child), _("Close") );
      gtk_clist_set_selection_mode (GTK_CLIST (GTK_FILE_SELECTION
                                               (pandora_fsb)->file_list),
                                    GTK_SELECTION_EXTENDED);
#else
      gtk_file_selection_set_select_multiple(GTK_FILE_SELECTION (pandora_fsb),
                                             TRUE);

      gtk_button_set_label(GTK_FILE_SELECTION (pandora_fsb)->ok_button,
                           _("Add"));
      gtk_button_set_label(GTK_FILE_SELECTION (pandora_fsb)->cancel_button,
                           _("Close"));
#endif

      /* Connect all the signals */
      g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (pandora_fsb)), "destroy",
                        G_CALLBACK (gtk_widget_destroyed),
                        G_OBJECT (pandora_fsb));

      g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (pandora_fsb)->ok_button),
                        "clicked",
                        G_CALLBACK (pandora_file_selection_ok),
                        G_OBJECT (pandora_fsb));

      g_signal_connect (G_OBJECT (GTK_FILE_SELECTION
                                  (pandora_fsb)->cancel_button),
                        "clicked",
                        G_CALLBACK (pandora_file_selection_cancel),
                        G_OBJECT (pandora_fsb));
      gimp_help_set_help_data (GTK_FILE_SELECTION (pandora_fsb)->cancel_button,
			       _("Click here to cancel file selection"), NULL);
    }

  if (!GTK_WIDGET_VISIBLE (pandora_fsb))
    gtk_widget_show (pandora_fsb);
}

static void
pandora_add_layer_mask (gint32 pan_img, gint32 layer,
                        gint32 w, gint32 h, gdouble overlap)
{
#ifndef GIMP_1_2
  if (gimp_gradients_set_gradient ("German_flag_smooth"))
#endif /* GIMP_1_2 */
  {
    gdouble hgrad = (gdouble)h / 2.;
    gint32 mask;

#ifdef GIMP_1_2
    guchar oldFGr, oldFGg, oldFGb;
    guchar oldBGr, oldBGg, oldBGb;
    gimp_palette_get_foreground (&oldFGr, &oldFGg, &oldFGb);
    gimp_palette_get_background (&oldBGr, &oldBGg, &oldBGb);
    gimp_palette_set_foreground (0xff, 0xff, 0xff);
    gimp_palette_set_background (   0,    0,    0);
#else /* GIMP_1_2 */
    GimpRGB black, white;
    GimpRGB oldFG, oldBG;
    gimp_palette_get_foreground (&oldFG);
    gimp_palette_get_background (&oldBG);
    gimp_rgb_set (&black, 0., 0., 0.);
    gimp_rgb_set (&white, 255., 255., 255.);
    gimp_palette_set_background (&white);
    gimp_palette_set_foreground (&black);
#endif /* GIMP_1_2 */

    /* Give it a layer mask */
    mask = gimp_layer_create_mask (layer, GIMP_WHITE_MASK);
    gimp_image_add_layer_mask (pan_img, layer, mask);

    /* might need GIMP_ADD_WHITE_MASK for 1.3 */
    gimp_blend (mask, GIMP_FG_BG_RGB, GIMP_NORMAL_MODE,
                GIMP_LINEAR, 100.0, 0.0, GIMP_REPEAT_NONE,
                FALSE, 0, 0.0,  /* supersampling stuff */
                w * overlap * .5, hgrad, 0.0, hgrad);

#ifdef GIMP_1_2
    gimp_palette_set_foreground (oldFGr, oldFGg, oldFGb);
    gimp_palette_set_background (oldBGr, oldBGg, oldBGb);
#else /* GIMP_1_2 */
    gimp_palette_set_background (&oldBG);
    gimp_palette_set_foreground (&oldFG);
#endif /* GIMP_1_2 */
  }
}

static void
pandora_ok_callback (GtkWidget *widget,
                     gpointer   data)
{
  int i;
  int pan_img = 0;
  int pan_img_w = 0, pan_img_h = 0;
  int pan_num_layers = 0;
  gint32 pan_layers[MAXIMAGES];
  int layerXoff[MAXIMAGES];
  int w = 0, h;

  layerXoff[0] = 0;

  /* Get the list of files */
  for (i=0; ; ++i)
  {
    gchar* fnam;
    gchar* bnam;
    int cur_img;
    int numlayers;
    int *layers;
    int j;

    if (gtk_clist_get_text(GTK_CLIST (pandora_filelist), i, 0, &fnam) == 0)
      break;
    bnam = basename(fnam);
    cur_img = gimp_file_load (GIMP_RUN_INTERACTIVE, fnam, bnam);
    /* should pass the real run mode */
    if (cur_img == -1)
      {
        printf("Couldn't open %s\n", fnam);
        continue;
      }

    gimp_image_set_filename (cur_img, bnam);
    w = gimp_image_width (cur_img);
    h = gimp_image_height (cur_img);
    if (pan_img)
      {
        /* Resize the image bigger if need be */
        if (w > pan_img_w)
          pan_img_w = w*2;
        if (h > pan_img_h)
          pan_img_h = h*2;
        gimp_image_resize(pan_img, pan_img_w, pan_img_h, 0, 0);
      }
    else
      {
        /* Make a new image, twice the size of this first image */
        pan_img_w = w*2;
        pan_img_h = h*2;
        pan_img = gimp_image_new(pan_img_w, pan_img_h, GIMP_RGB_IMAGE);
      }

    /* Now it's safe, so stick the image in */
    layers = gimp_image_get_layers(cur_img, &numlayers);
    for (j = 0; j < numlayers && pan_num_layers < MAXIMAGES; ++j)
      {
        /* Make a new layer in pan_img and paste the image's layer into it. */
#ifdef GIMP_1_2
        pan_layers[pan_num_layers] = gimp_layer_new(pan_img, bnam, w, h,
                                                    GIMP_RGBA_IMAGE, 100.0,
                                                    GIMP_NORMAL_MODE);
        gimp_image_add_layer(pan_img, pan_layers[pan_num_layers], 0);
        gimp_edit_copy(layers[j]);
        gimp_edit_paste(pan_layers[pan_num_layers], FALSE);
#else
        /* gimp 1.3 has this great new call: */
        pan_layers[pan_num_layers] = gimp_layer_new_from_drawable(layers[j]);
        gimp_image_add_layer(pan_img, pan_layers[pan_num_layers], 0);
#endif /* GIMP_1.2 */

        /* Give it a layer mask with a gradient on the left,
         * except the first image.
         * Bah, APIs changed a lot for this part!
         */
        if (pan_num_layers > 0)
          pandora_add_layer_mask (pan_img, pan_layers[pan_num_layers],
                                  w, h, avals.overlap);

        /* Calculate the offset of the next image
         * based on the size of this one.
         * Don't try to set it now -- the image isn't big enough yet.
         */
        layerXoff[pan_num_layers+1] = layerXoff[pan_num_layers]
          + (w * (1. - avals.overlap));
        ++pan_num_layers;
      }

    /* Now we're done with the temp image. */
    gimp_image_delete(cur_img);

    if (pan_num_layers >= MAXIMAGES)
      break;

  } /* end of loop over files */

  /* Now we're probably left with a floating selection from the
   * last paste, and we need to get rid of it somehow.
   */
  i = gimp_image_floating_selection (pan_img);
  if (i != -1)
    gimp_floating_sel_anchor (i);

  /* Now, resize the image to the right size, and move all the layers.
   * Allow some extra room on the right in case the user was conservative
   * with estimating overlap.
   */
  printf("Final image should have %d layers and width %d\n",
         pan_num_layers, layerXoff[pan_num_layers]);
  gimp_image_resize(pan_img,
                    layerXoff[pan_num_layers] + w * .6, pan_img_h,
                    0, 0);

  /* Loop over all layers in the final image,
   * moving them to their final resting place.
   */
  for (i=0; i<pan_num_layers; ++i)
    {
      GimpDrawable* drawable;

      gimp_layer_translate (pan_layers[i], layerXoff[i],
                            (pan_img_h -
                             gimp_drawable_height (pan_layers[i])) / 2);

      /* Try to flush everything.
         Is all this really needed?  I have no idea!
      */
      drawable = gimp_drawable_get (pan_layers[i]);
      gimp_drawable_flush (drawable);
      //gimp_drawable_merge_shadow (pan_layers[i], TRUE);
      //gimp_drawable_update (pan_layers[i], 0, 0, new_width, new_height);
      gimp_drawable_detach (drawable);
    }

  gimp_display_new (pan_img);
  gimp_displays_flush ();

  aint.run = TRUE;
  gtk_widget_destroy (GTK_WIDGET (data));
}

static gint
pandora_dialog (void)
{
  GtkWidget *dlg;
  GtkWidget *button;
  GtkWidget *hbox;
  GtkWidget *logo_box;
  GtkWidget *frame;
  GtkWidget *table;
  GtkWidget *main_vbox;
  GtkWidget *vbox;
  GtkWidget *scrolled_win;

  GtkObject *scale;

#ifndef GIMP_1_2
  GdkPixbuf* icon_pixbuf;
  GtkWidget *preview;
#endif

  gimp_ui_init ("pandora", TRUE);

  dlg = gimp_dialog_new (_("Pandora"), "pandora",
			 gimp_standard_help_func, "filters/pandora.html",
			 GTK_WIN_POS_MOUSE,
			 FALSE, TRUE, FALSE,

			 GTK_STOCK_CANCEL, gtk_widget_destroy,
			 NULL, 1, NULL, FALSE, TRUE,

			 GTK_STOCK_OK, pandora_ok_callback,
			 NULL, NULL, NULL, TRUE, FALSE,

			 NULL);

  g_signal_connect (G_OBJECT (dlg), "destroy",
                    G_CALLBACK (gtk_main_quit),
                    NULL);

  hbox = gtk_hbox_new (FALSE, 6);
  gtk_container_set_border_width (GTK_CONTAINER (hbox), 6);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->vbox), hbox, TRUE, TRUE, 0);

  logo_box = gtk_vbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), logo_box, FALSE, FALSE, 0);

  frame = gtk_frame_new (NULL);
  gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
  gtk_box_pack_start (GTK_BOX (logo_box), frame, FALSE, FALSE, 0);

#ifdef DO_ICON_HERE_IF_THERE_IS_ONE
  icon_pixbuf = gdk_pixbuf_new_from_inline(-1, pandora_icon, FALSE, NULL);

  preview = gtk_image_new_from_pixbuf (icon_pixbuf);

  gtk_container_add (GTK_CONTAINER (frame), preview);
  gtk_widget_show (preview);
#endif /* ! GIMP_1_2 */
  gtk_widget_show (frame);
  gtk_widget_show (logo_box);

  main_vbox = gtk_vbox_new (FALSE, 4);
  gtk_container_add (GTK_CONTAINER (hbox), main_vbox);

  frame = gtk_frame_new (_("Panorama Files"));
  gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_ETCHED_IN);
  gtk_box_pack_start (GTK_BOX (main_vbox), frame, FALSE, FALSE, 0);

  vbox = gtk_vbox_new (FALSE, 4);
  gtk_container_add (GTK_CONTAINER (frame), vbox);

  pandora_filelist = gtk_clist_new (1);
  scrolled_win = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_win),
                                  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_clist_set_selection_mode (GTK_CLIST (pandora_filelist),
                                GTK_SELECTION_BROWSE);
  gtk_widget_set_usize (pandora_filelist, FILELIST_WIDTH, FILELIST_HEIGHT);
  /*
  gtk_signal_connect (GTK_OBJECT (pandora_filelist), "select_row",
                      (GtkSignalFunc) procedure_select_callback,
                      dbbrowser);
  */
  gtk_box_pack_start (GTK_BOX (vbox), pandora_filelist, TRUE, TRUE, 4);
  gtk_widget_show (pandora_filelist);

  button = gtk_button_new_with_label (_("Choose Files..."));
  gtk_box_pack_start (GTK_BOX (vbox), button, TRUE, TRUE, 0);
  gtk_widget_show (button);
  g_signal_connect (G_OBJECT (button), "clicked",
                    G_CALLBACK (pandora_show_file_selection), NULL);

  table = gtk_table_new (1, 1, FALSE);
  //gtk_table_set_col_spacings (GTK_TABLE (table2), 4);
  //gtk_table_set_row_spacings (GTK_TABLE (table2), 2);
  gtk_container_set_border_width (GTK_CONTAINER (table), 4);
  gtk_box_pack_start (GTK_BOX (main_vbox), table, TRUE, TRUE, 0);

  scale = gimp_scale_entry_new (GTK_TABLE (table), 0, 0,
                                _("Overlap:"), SCALE_WIDTH, SPIN_WIDTH,
                                avals.overlap, 0.0, 1.0, 0.01, 0.1, 3,
                                TRUE, 0, 0,
                                //_("How much each image overlaps the last"),
                                // Why do tooltips always give a Gtk-CRITICAL?
                                NULL,
                                NULL);
  g_signal_connect (G_OBJECT (scale), "value_changed",
                    G_CALLBACK (gimp_double_adjustment_update),
                    &avals.overlap);

  gtk_widget_show (vbox);
  gtk_widget_show (table);
  gtk_widget_show (frame);
  gtk_widget_show (main_vbox);
  gtk_widget_show (hbox);
  gtk_widget_show (dlg);

  /* Show the fsb by default -- without it we can do nothing */
  pandora_show_file_selection (button);

  gtk_main ();
  gdk_flush ();

  return aint.run;
}

gint32 make_pandora ()
{
  return -1;
}

